import typing
from datetime import datetime

from typing import List, Optional

from pydantic.main import BaseModel

from traveltimepy import (
    Coordinates,
    Range,
    PublicTransport,
    Driving,
    Ferry,
    Walking,
    Cycling,
    DrivingTrain,
    CyclingPublicTransport,
    LevelOfDetail,
)
from traveltimepy.dto.common import PolygonsFilter, RenderMode, Snapping
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.responses.time_map import TimeMapResponse
from traveltimepy.itertools import split, flatten


class DepartureSearch(BaseModel):
    id: str
    coords: Coordinates
    departure_time: datetime
    travel_time: int
    transportation: typing.Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ]
    range: Optional[Range]
    level_of_detail: Optional[LevelOfDetail]
    snapping: Optional[Snapping]
    polygons_filter: Optional[PolygonsFilter]
    remove_water_bodies: Optional[bool]
    render_mode: Optional[RenderMode]


class ArrivalSearch(BaseModel):
    id: str
    coords: Coordinates
    arrival_time: datetime
    travel_time: int
    transportation: typing.Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ]
    range: Optional[Range]
    level_of_detail: Optional[LevelOfDetail]
    snapping: Optional[Snapping]
    polygons_filter: Optional[PolygonsFilter]
    remove_water_bodies: Optional[bool]
    render_mode: Optional[RenderMode]


class Intersection(BaseModel):
    id: str
    search_ids: List[str]


class Union(BaseModel):
    id: str
    search_ids: List[str]


class TimeMapRequest(TravelTimeRequest[TimeMapResponse]):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]
    unions: List[Union]
    intersections: List[Intersection]

    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        return [
            TimeMapRequest(
                departure_searches=departures,
                arrival_searches=arrivals,
                unions=self.unions,
                intersections=self.intersections,
            )
            for departures, arrivals in split(
                self.departure_searches, self.arrival_searches, window_size
            )
        ]

    def merge(self, responses: List[TimeMapResponse]) -> TimeMapResponse:
        if len(self.unions) != 0:
            return TimeMapResponse(
                results=list(
                    filter(
                        lambda res: res.search_id == "Union search",
                        flatten([response.results for response in responses]),
                    )
                )
            )
        elif len(self.intersections) != 0:
            return TimeMapResponse(
                results=list(
                    filter(
                        lambda res: res.search_id == "Intersection search",
                        flatten([response.results for response in responses]),
                    )
                )
            )
        else:
            return TimeMapResponse(
                results=sorted(
                    flatten([response.results for response in responses]),
                    key=lambda res: res.search_id,
                )
            )



from shapely.geometry import (
    Point,
    LineString,
    Polygon,
    MultiPoint,
    MultiLineString,
    MultiPolygon,
)

SUPPORTED_GEOMETRY_TYPES = [
    Point,
    LineString,
    Polygon,
    MultiPoint,
    MultiLineString,
    MultiPolygon,
]



from typing import List, Optional

from pydantic.main import BaseModel


class Properties(BaseModel):
    min: Optional[int] = None
    max: Optional[int] = None
    mean: Optional[int] = None


class Cell(BaseModel):
    id: str
    properties: Properties


class GeohashResult(BaseModel):
    search_id: str
    cells: List[Cell]


class GeohashResponse(BaseModel):
    results: List[GeohashResult]



import asyncio
import time
from datetime import datetime

from benchmarks.common import generate_coordinates
from traveltimepy import TravelTimeSdk, Driving


async def generate_isochrones(size: int):
    sdk = TravelTimeSdk("APP_ID", "API_KEY")
    coordinates = generate_coordinates(51.507609, -0.128315, 0.05, size)
    return await sdk.time_map_async(
        coordinates=coordinates, transportation=Driving(), arrival_time=datetime.now()
    )


if __name__ == "__main__":
    start = time.perf_counter()
    response = asyncio.run(generate_isochrones(50))
    request_time = time.perf_counter() - start
    print("Request completed in {0:.0f}s for 50 isochrones".format(request_time))

    start = time.perf_counter()
    response2 = asyncio.run(generate_isochrones(100))
    request_time = time.perf_counter() - start
    print("Request completed in {0:.0f}s for 100 isochrones".format(request_time))

    start = time.perf_counter()
    response3 = asyncio.run(generate_isochrones(300))
    request_time = time.perf_counter() - start
    print("Request completed in {0:.0f}s for 300 isochrones".format(request_time))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: RequestsCommon.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14RequestsCommon.proto\x12)com.igeolise.traveltime.rabbitmq.requests\"\"\n\x06\x43oords\x12\x0b\n\x03lat\x18\x01 \x01(\x02\x12\x0b\n\x03lng\x18\x02 \x01(\x02\"\xc6\x02\n\x0eTransportation\x12K\n\x04type\x18\x01 \x01(\x0e\x32=.com.igeolise.traveltime.rabbitmq.requests.TransportationType\x12\\\n\x0fpublicTransport\x18\x02 \x01(\x0b\x32\x41.com.igeolise.traveltime.rabbitmq.requests.PublicTransportDetailsH\x00\x12p\n\x19\x64rivingAndPublicTransport\x18\x03 \x01(\x0b\x32K.com.igeolise.traveltime.rabbitmq.requests.DrivingAndPublicTransportDetailsH\x00\x42\x17\n\x15transportationDetails\"6\n\x16PublicTransportDetails\x12\x1c\n\x14walkingTimeToStation\x18\x01 \x01(\r\"s\n DrivingAndPublicTransportDetails\x12\x1c\n\x14walkingTimeToStation\x18\x01 \x01(\r\x12\x1c\n\x14\x64rivingTimeToStation\x18\x02 \x01(\r\x12\x13\n\x0bparkingTime\x18\x03 \x01(\x11*\xb8\x01\n\x12TransportationType\x12\x14\n\x10PUBLIC_TRANSPORT\x10\x00\x12\x0b\n\x07\x44RIVING\x10\x01\x12 \n\x1c\x44RIVING_AND_PUBLIC_TRANSPORT\x10\x02\x12\x15\n\x11\x44RIVING_AND_FERRY\x10\x03\x12\x0b\n\x07WALKING\x10\x04\x12\x15\n\x11WALKING_AND_FERRY\x10\x07\x12\x0b\n\x07\x43YCLING\x10\x05\x12\x15\n\x11\x43YCLING_AND_FERRY\x10\x06*!\n\nTimePeriod\x12\x13\n\x0fWEEKDAY_MORNING\x10\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'RequestsCommon_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _TRANSPORTATIONTYPE._serialized_start=606
  _TRANSPORTATIONTYPE._serialized_end=790
  _TIMEPERIOD._serialized_start=792
  _TIMEPERIOD._serialized_end=825
  _COORDS._serialized_start=67
  _COORDS._serialized_end=101
  _TRANSPORTATION._serialized_start=104
  _TRANSPORTATION._serialized_end=430
  _PUBLICTRANSPORTDETAILS._serialized_start=432
  _PUBLICTRANSPORTDETAILS._serialized_end=486
  _DRIVINGANDPUBLICTRANSPORTDETAILS._serialized_start=488
  _DRIVINGANDPUBLICTRANSPORTDETAILS._serialized_end=603
# @@protoc_insertion_point(module_scope)



from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Optional


@dataclass
class TransportationInfo:
    code: int
    name: str


class ProtoTransportation(Enum):
    PUBLIC_TRANSPORT = TransportationInfo(0, "pt")
    DRIVING = TransportationInfo(1, "driving")
    DRIVING_AND_PUBLIC_TRANSPORT = TransportationInfo(2, "pt")
    DRIVING_FERRY = TransportationInfo(3, "driving+ferry")
    WALKING = TransportationInfo(4, "walking")
    CYCLING = TransportationInfo(5, "cycling")
    CYCLING_FERRY = TransportationInfo(6, "cycling+ferry")
    WALKING_FERRY = TransportationInfo(7, "walking+ferry")


@dataclass
class PublicTransportWithDetails:

    walking_time_to_station: Optional[int] = None
    """Limit on walking path duration. Must be <= 1800"""

    TYPE: ClassVar[ProtoTransportation] = ProtoTransportation.PUBLIC_TRANSPORT


@dataclass
class DrivingAndPublicTransportWithDetails:

    walking_time_to_station: Optional[int] = None
    """Limit on walking path duration. Must be <= 1800"""

    driving_time_to_station: Optional[int] = None
    """Limit on driving path duration. Must be <= 1800"""

    parking_time: Optional[int] = None
    """
    Constant penalty to simulate finding a parking spot in seconds.
    Must be less than the overall travel time limit.
    """

    TYPE: ClassVar[ProtoTransportation] = (
        ProtoTransportation.DRIVING_AND_PUBLIC_TRANSPORT
    )


class ProtoCountry(str, Enum):
    NETHERLANDS = "nl"
    AUSTRIA = "at"
    UNITED_KINGDOM = "uk"
    BELGIUM = "be"
    GERMANY = "de"
    FRANCE = "fr"
    IRELAND = "ie"
    LITHUANIA = "lt"
    UNITED_STATES = "us"
    SOUTH_AFRICA = "za"
    ROMANIA = "ro"
    PORTUGAL = "pt"
    PHILIPPINES = "ph"
    NEW_ZEALAND = "nz"
    NORWAY = "no"
    LATVIA = "lv"
    JAPAN = "jp"
    INDIA = "in"
    INDONESIA = "id"
    HUNGARY = "hu"
    GREECE = "gr"
    FINLAND = "fi"
    DENMARK = "dk"
    CANADA = "ca"
    AUSTRALIA = "au"
    SINGAPORE = "sg"
    SWITZERLAND = "ch"
    SPAIN = "es"
    ITALY = "it"
    POLAND = "pl"
    SWEDEN = "se"
    LIECHTENSTEIN = "li"
    MEXICO = "mx"
    SAUDI_ARABIA = "sa"
    SERBIA = "rs"
    SLOVENIA = "si"



import pytest  # noqa

from traveltimepy import Coordinates
from traveltimepy.wkt import (
    PointModel,
    LineStringModel,
    PolygonModel,
    MultiPointModel,
    MultiLineStringModel,
    MultiPolygonModel,
)

# Updated mock data for LineStringModel with two distinct points
mock_point1 = PointModel(coordinates=Coordinates(lat=10, lng=20))
mock_point2 = PointModel(coordinates=Coordinates(lat=30, lng=40))
mock_linestring = LineStringModel(coordinates=[mock_point1, mock_point2])
mock_polygon = PolygonModel(exterior=mock_linestring, interiors=[mock_linestring])
mock_multipoint = MultiPointModel(coordinates=[mock_point1, mock_point2])
mock_multilinestring = MultiLineStringModel(
    coordinates=[mock_linestring, mock_linestring]
)
mock_multipolygon = MultiPolygonModel(coordinates=[mock_polygon])


# Test functions
def test_pointmodel_pretty_print(capsys):
    mock_point1.pretty_print()
    captured = capsys.readouterr()
    assert captured.out == "POINT: 10.0, 20.0\n"


def test_linestringmodel_pretty_print(capsys):
    mock_linestring.pretty_print()
    captured = capsys.readouterr()
    assert captured.out == "LINE STRING:\n\tPOINT: 10.0, 20.0\n\tPOINT: 30.0, 40.0\n"


def test_polygonmodel_pretty_print(capsys):
    mock_polygon.pretty_print()
    captured = capsys.readouterr()
    assert captured.out == (
        "POLYGON:\n\tEXTERIOR:\n\t\tLINE STRING:\n\t\t\tPOINT: 10.0, 20.0\n\t\t\tPOINT: 30.0, 40.0\n"
        "\tINTERIORS:\n\t\tLINE STRING:\n\t\t\tPOINT: 10.0, 20.0\n\t\t\tPOINT: 30.0, 40.0\n"
    )


def test_multipointmodel_pretty_print(capsys):
    mock_multipoint.pretty_print()
    captured = capsys.readouterr()
    assert captured.out == "MULTIPOINT:\n\tPOINT: 10.0, 20.0\n\tPOINT: 30.0, 40.0\n"


def test_multilinestringmodel_pretty_print(capsys):
    mock_multilinestring.pretty_print()
    captured = capsys.readouterr()
    assert captured.out == (
        "MULTILINESTRING:\n\tLINE STRING:\n\t\tPOINT: 10.0, 20.0\n\t\tPOINT: 30.0, 40.0\n"
        "\tLINE STRING:\n\t\tPOINT: 10.0, 20.0\n\t\tPOINT: 30.0, 40.0\n"
    )


def test_multipolygonmodel_pretty_print(capsys):
    mock_multipolygon.pretty_print()
    captured = capsys.readouterr()
    assert captured.out == (
        "MULTIPOLYGON:\n\tPOLYGON:\n\t\tEXTERIOR:\n\t\t\tLINE STRING:\n\t\t\t\tPOINT: 10.0, 20.0\n\t\t\t\tPOINT: 30.0, 40.0\n"
        "\t\tINTERIORS:\n\t\t\tLINE STRING:\n\t\t\t\tPOINT: 10.0, 20.0\n\t\t\t\tPOINT: 30.0, 40.0\n"
    )



import pytest
from datetime import datetime

from traveltimepy import Coordinates, Driving, LevelOfDetail, Range, TravelTimeSdk


@pytest.mark.asyncio
async def test_departures(sdk: TravelTimeSdk):
    results = await sdk.time_map_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        departure_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
        level_of_detail=LevelOfDetail(scale_type="simple", level="lowest"),
    )
    assert len(results) == 2


@pytest.mark.asyncio
async def test_departures_geojson(sdk: TravelTimeSdk):
    results = await sdk.time_map_geojson_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        departure_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
        level_of_detail=LevelOfDetail(scale_type="simple", level="lowest"),
    )
    assert len(results.features) == 2


@pytest.mark.asyncio
async def test_departures_wkt(sdk: TravelTimeSdk):
    response = await sdk.time_map_wkt_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        departure_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
        level_of_detail=LevelOfDetail(scale_type="simple", level="lowest"),
    )
    assert len(response.results) == 2


@pytest.mark.asyncio
async def test_departures_wkt_no_holes(sdk: TravelTimeSdk):
    response = await sdk.time_map_wkt_no_holes_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        departure_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
        level_of_detail=LevelOfDetail(scale_type="simple", level="lowest"),
    )
    assert len(response.results) == 2


@pytest.mark.asyncio
async def test_arrivals(sdk: TravelTimeSdk):
    results = await sdk.time_map_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        arrival_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
        level_of_detail=LevelOfDetail(scale_type="simple", level="lowest"),
    )
    assert len(results) == 2


@pytest.mark.asyncio
async def test_arrivals_geojson(sdk: TravelTimeSdk):
    results = await sdk.time_map_geojson_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        arrival_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
        level_of_detail=LevelOfDetail(scale_type="simple", level="lowest"),
    )
    assert len(results.features) == 2


@pytest.mark.asyncio
async def test_arrivals_wkt(sdk: TravelTimeSdk):
    response = await sdk.time_map_wkt_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        arrival_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
        level_of_detail=LevelOfDetail(scale_type="simple", level="lowest"),
    )
    assert len(response.results) == 2


@pytest.mark.asyncio
async def test_arrivals_wkt_no_holes(sdk: TravelTimeSdk):
    response = await sdk.time_map_wkt_no_holes_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        arrival_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
        level_of_detail=LevelOfDetail(scale_type="simple", level="lowest"),
    )
    assert len(response.results) == 2


@pytest.mark.asyncio
async def test_union_departures(sdk: TravelTimeSdk):
    result = await sdk.time_map_union_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        departure_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
        level_of_detail=LevelOfDetail(scale_type="simple", level="lowest"),
    )
    assert len(result.shapes) > 0


@pytest.mark.asyncio
async def test_intersection_arrivals(sdk: TravelTimeSdk):
    result = await sdk.time_map_intersection_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            Coordinates(lat=51.517609, lng=-0.138315),
        ],
        arrival_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
        level_of_detail=LevelOfDetail(scale_type="simple", level="lowest"),
    )
    assert len(result.shapes) > 0



from typing import List

from geojson_pydantic import FeatureCollection
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.requests.time_map import (
    DepartureSearch,
    ArrivalSearch,
)
from traveltimepy.itertools import split, flatten


class TimeMapRequestGeojson(TravelTimeRequest[FeatureCollection]):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]

    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        return [
            TimeMapRequestGeojson(
                departure_searches=departures,
                arrival_searches=arrivals,
            )
            for departures, arrivals in split(
                self.departure_searches, self.arrival_searches, window_size
            )
        ]

    def merge(self, responses: List[FeatureCollection]) -> FeatureCollection:
        merged_features = flatten([response.features for response in responses])
        return FeatureCollection(type="FeatureCollection", features=merged_features)



import pytest

from traveltimepy import Coordinates
from traveltimepy.dto.common import PropertyProto
from traveltimepy.dto.requests.time_filter_proto import (
    DrivingAndPublicTransportWithDetails,
    ProtoTransportation,
    ProtoCountry,
    PublicTransportWithDetails,
)
from traveltimepy.sdk import TravelTimeSdk


@pytest.mark.asyncio
async def test_one_to_many(proto_sdk: TravelTimeSdk):
    results = await proto_sdk.time_filter_proto_async(
        origin=Coordinates(lat=51.425709, lng=-0.122061),
        destinations=[
            Coordinates(lat=51.348605, lng=-0.314783),
            Coordinates(lat=51.337205, lng=-0.315793),
        ],
        transportation=ProtoTransportation.DRIVING_FERRY,
        travel_time=7200,
        country=ProtoCountry.UNITED_KINGDOM,
    )
    assert len(results.travel_times) == 2 and len(results.distances) == 0


@pytest.mark.asyncio
async def test_many_to_one(proto_sdk: TravelTimeSdk):
    results = await proto_sdk.time_filter_proto_async(
        origin=Coordinates(lat=51.425709, lng=-0.122061),
        destinations=[
            Coordinates(lat=51.348605, lng=-0.314783),
            Coordinates(lat=51.337205, lng=-0.315793),
        ],
        transportation=ProtoTransportation.DRIVING_FERRY,
        travel_time=7200,
        country=ProtoCountry.UNITED_KINGDOM,
        one_to_many=False,
    )
    assert len(results.travel_times) == 2 and len(results.distances) == 0


@pytest.mark.asyncio
async def test_one_to_many_with_distances(proto_sdk: TravelTimeSdk):
    results = await proto_sdk.time_filter_proto_async(
        origin=Coordinates(lat=51.425709, lng=-0.122061),
        destinations=[
            Coordinates(lat=51.348605, lng=-0.314783),
            Coordinates(lat=51.337205, lng=-0.315793),
        ],
        transportation=ProtoTransportation.DRIVING_FERRY,
        travel_time=7200,
        country=ProtoCountry.UNITED_KINGDOM,
        properties=[PropertyProto.DISTANCE],
    )
    assert len(results.travel_times) == 2 and len(results.distances) == 2


@pytest.mark.asyncio
async def test_many_to_one_with_distances(proto_sdk: TravelTimeSdk):
    results = await proto_sdk.time_filter_proto_async(
        origin=Coordinates(lat=51.425709, lng=-0.122061),
        destinations=[
            Coordinates(lat=51.348605, lng=-0.314783),
            Coordinates(lat=51.337205, lng=-0.315793),
        ],
        transportation=ProtoTransportation.DRIVING_FERRY,
        travel_time=7200,
        country=ProtoCountry.UNITED_KINGDOM,
        one_to_many=False,
        properties=[PropertyProto.DISTANCE],
    )
    assert len(results.travel_times) == 2 and len(results.distances) == 2


async def test_one_to_many_pt_with_params(proto_sdk: TravelTimeSdk):
    results = await proto_sdk.time_filter_proto_async(
        origin=Coordinates(lat=51.425709, lng=-0.122061),
        destinations=[
            Coordinates(lat=51.348605, lng=-0.314783),
            Coordinates(lat=51.337205, lng=-0.315793),
        ],
        transportation=PublicTransportWithDetails(walking_time_to_station=900),
        travel_time=7200,
        country=ProtoCountry.UNITED_KINGDOM,
    )
    assert len(results.travel_times) == 2 and len(results.distances) == 0


@pytest.mark.asyncio
async def test_many_to_one_pt_with_params(proto_sdk: TravelTimeSdk):
    results = await proto_sdk.time_filter_proto_async(
        origin=Coordinates(lat=51.425709, lng=-0.122061),
        destinations=[
            Coordinates(lat=51.348605, lng=-0.314783),
            Coordinates(lat=51.337205, lng=-0.315793),
        ],
        transportation=PublicTransportWithDetails(walking_time_to_station=900),
        travel_time=7200,
        country=ProtoCountry.UNITED_KINGDOM,
        one_to_many=False,
    )
    assert len(results.travel_times) == 2 and len(results.distances) == 0


async def test_one_to_many_driving_and_pt_with_params(proto_sdk: TravelTimeSdk):
    results = await proto_sdk.time_filter_proto_async(
        origin=Coordinates(lat=51.425709, lng=-0.122061),
        destinations=[
            Coordinates(lat=51.348605, lng=-0.314783),
            Coordinates(lat=51.337205, lng=-0.315793),
        ],
        transportation=DrivingAndPublicTransportWithDetails(
            walking_time_to_station=900, driving_time_to_station=1800, parking_time=300
        ),
        travel_time=7200,
        country=ProtoCountry.UNITED_KINGDOM,
    )
    assert len(results.travel_times) == 2 and len(results.distances) == 0


@pytest.mark.asyncio
async def test_many_to_one_driving_and_pt_with_params(proto_sdk: TravelTimeSdk):
    results = await proto_sdk.time_filter_proto_async(
        origin=Coordinates(lat=51.425709, lng=-0.122061),
        destinations=[
            Coordinates(lat=51.348605, lng=-0.314783),
            Coordinates(lat=51.337205, lng=-0.315793),
        ],
        transportation=DrivingAndPublicTransportWithDetails(
            walking_time_to_station=900, driving_time_to_station=1800, parking_time=300
        ),
        travel_time=7200,
        country=ProtoCountry.UNITED_KINGDOM,
        one_to_many=False,
    )
    assert len(results.travel_times) == 2 and len(results.distances) == 0



from datetime import datetime
from typing import List, Optional, Dict, Union

from traveltimepy.dto.common import (
    CellProperty,
    GeohashCentroid,
    H3Centroid,
    Location,
    Coordinates,
    PolygonsFilter,
    Rectangle,
    Property,
    FullRange,
    Range,
    LevelOfDetail,
    PropertyProto,
    DepartureTime,
    ArrivalTime,
    RenderMode,
    Snapping,
)
from traveltimepy.dto.requests import geohash_fast, h3_fast, time_map_fast
from traveltimepy.dto.responses.geohash import GeohashResponse, GeohashResult
from traveltimepy.dto.responses.h3 import H3Response, H3Result
from traveltimepy.dto.responses.time_map_wkt import (
    TimeMapWKTResponse,
)
from traveltimepy.dto.responses.time_filter_proto import TimeFilterProtoResponse
from traveltimepy.dto.transportation import (
    PublicTransport,
    Driving,
    Ferry,
    Walking,
    Cycling,
    DrivingTrain,
    CyclingPublicTransport,
)
from traveltimepy.dto.requests.postcodes_zones import ZonesProperty
from traveltimepy.dto.requests.time_filter_proto import (
    DrivingAndPublicTransportWithDetails,
    ProtoCountry,
    ProtoTransportation,
    PublicTransportWithDetails,
)
from traveltimepy.dto.requests.time_filter_fast import Transportation
from traveltimepy.errors import ApiError

from traveltimepy import __version__
from traveltimepy.accept_type import AcceptType
from traveltimepy.itertools import join_opt
from traveltimepy.dto.requests.supported_locations import SupportedLocationsRequest

from traveltimepy.dto.responses.map_info import MapInfoResponse, Map
from traveltimepy.dto.responses.postcodes import PostcodesResponse, PostcodesResult
from traveltimepy.dto.responses.routes import RoutesResponse, RoutesResult
from traveltimepy.dto.responses.supported_locations import SupportedLocationsResponse
from traveltimepy.dto.responses.time_filter import TimeFilterResponse, TimeFilterResult
from traveltimepy.dto.responses.time_filter_fast import (
    TimeFilterFastResponse,
    TimeFilterFastResult,
)
from traveltimepy.dto.responses.time_map import TimeMapResponse, TimeMapResult
from traveltimepy.dto.responses.zones import (
    PostcodesDistrictsResponse,
    PostcodesSectorsResponse,
    PostcodesDistrictsResult,
    PostcodesSectorsResult,
)

from traveltimepy.mapper import (
    create_distance_map,
    create_geohash,
    create_geohash_fast,
    create_geohash_intersection,
    create_geohash_union,
    create_h3,
    create_h3_fast,
    create_h3_intersection,
    create_h3_union,
    create_time_filter,
    create_time_filter_fast,
    create_postcodes,
    create_districts,
    create_sectors,
    create_routes,
    create_proto_request,
    create_time_map,
    create_time_map_intersection,
    create_time_map_fast,
    create_time_map_fast_geojson,
    create_time_map_fast_wkt,
    create_time_map_union,
    create_time_map_geojson,
    create_time_map_wkt,
)

from traveltimepy.proto_http import send_proto_async
from traveltimepy.http import (
    send_get_async,
    send_post_async,
    SdkParams,
)

from geojson_pydantic import FeatureCollection


class TravelTimeSdk:
    def __init__(
        self,
        app_id: str,
        api_key: str,
        limit_per_host: int = 2,
        rate_limit: int = 60,
        time_window: int = 60,
        retry_attempts: int = 2,
        host: str = "api.traveltimeapp.com",
        proto_host: str = "proto.api.traveltimeapp.com",
        timeout: int = 300,
        user_agent: Optional[str] = None,
    ) -> None:
        self._app_id = app_id
        self._api_key = api_key

        if user_agent is not None:
            self._user_agent = user_agent
        else:
            self._user_agent = f"Travel Time Python SDK {__version__}"

        self._sdk_params = SdkParams(
            host,
            proto_host,
            limit_per_host,
            rate_limit,
            time_window,
            retry_attempts,
            timeout,
        )

    async def time_filter_async(
        self,
        locations: List[Location],
        search_ids: Dict[str, List[str]],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        properties: Optional[List[Property]] = None,
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        range: Optional[FullRange] = None,
        snapping: Optional[Snapping] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[TimeFilterResult]:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            TimeFilterResponse,
            v4_endpoint_path or "time-filter",
            self._headers(AcceptType.JSON),
            create_time_filter(
                locations,
                search_ids,
                transportation,
                properties,
                time_info,
                travel_time,
                range,
                snapping,
            ),
            self._sdk_params,
        )

        return resp.results

    async def map_info_async(
        self,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[Map]:
        res = await send_get_async(
            MapInfoResponse,
            v4_endpoint_path or "map-info",
            self._headers(AcceptType.JSON),
            self._sdk_params,
            None,
        )
        return res.maps

    async def geocoding_async(
        self,
        query: str,
        limit: Optional[int] = None,
        within_countries: Optional[List[str]] = None,
        format_name: Optional[bool] = None,
        format_exclude_country: Optional[bool] = None,
        bounds: Optional[Rectangle] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> FeatureCollection:
        return await send_get_async(
            FeatureCollection,
            v4_endpoint_path or "geocoding/search",
            self._headers(AcceptType.JSON),
            self._sdk_params,
            self._geocoding_params(
                query,
                limit,
                within_countries,
                format_name,
                format_exclude_country,
                bounds,
            ),
        )

    async def geocoding_reverse_async(
        self,
        lat: float,
        lng: float,
        v4_endpoint_path: Optional[str] = None,
    ) -> FeatureCollection:
        return await send_get_async(
            FeatureCollection,
            v4_endpoint_path or "geocoding/reverse",
            self._headers(AcceptType.JSON),
            self._sdk_params,
            self._geocoding_reverse_params(lat, lng),
        )

    async def supported_locations_async(
        self,
        locations: List[Location],
        v4_endpoint_path: Optional[str] = None,
    ) -> SupportedLocationsResponse:
        return await send_post_async(
            SupportedLocationsResponse,
            v4_endpoint_path or "supported-locations",
            self._headers(AcceptType.JSON),
            SupportedLocationsRequest(locations=locations),
            self._sdk_params,
        )

    async def time_map_fast_async(
        self,
        coordinates: List[Coordinates],
        transportation: time_map_fast.Transportation,
        travel_time: int = 3600,
        one_to_many: bool = True,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        render_mode: Optional[RenderMode] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[TimeMapResult]:
        resp = await send_post_async(
            TimeMapResponse,
            v4_endpoint_path or "time-map/fast",
            self._headers(AcceptType.JSON),
            create_time_map_fast(
                coordinates,
                transportation,
                travel_time,
                level_of_detail,
                snapping,
                polygons_filter,
                render_mode,
                one_to_many,
            ),
            self._sdk_params,
        )
        return resp.results

    async def time_map_fast_geojson_async(
        self,
        coordinates: List[Coordinates],
        transportation: time_map_fast.Transportation,
        travel_time: int = 3600,
        one_to_many: bool = True,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        render_mode: Optional[RenderMode] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> FeatureCollection:
        resp = await send_post_async(
            FeatureCollection,
            v4_endpoint_path or "time-map/fast",
            self._headers(AcceptType.GEO_JSON),
            create_time_map_fast_geojson(
                coordinates,
                transportation,
                travel_time,
                level_of_detail,
                snapping,
                polygons_filter,
                render_mode,
                one_to_many,
            ),
            self._sdk_params,
        )
        return resp

    async def time_map_fast_wkt_async(
        self,
        coordinates: List[Coordinates],
        transportation: time_map_fast.Transportation,
        travel_time: int = 3600,
        one_to_many: bool = True,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        render_mode: Optional[RenderMode] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> TimeMapWKTResponse:
        resp = await send_post_async(
            TimeMapWKTResponse,
            v4_endpoint_path or "time-map/fast",
            self._headers(AcceptType.WKT),
            create_time_map_fast_wkt(
                coordinates,
                transportation,
                travel_time,
                level_of_detail,
                snapping,
                polygons_filter,
                render_mode,
                one_to_many,
            ),
            self._sdk_params,
        )
        return resp

    async def time_map_fast_wkt_no_holes_async(
        self,
        coordinates: List[Coordinates],
        transportation: time_map_fast.Transportation,
        travel_time: int = 3600,
        one_to_many: bool = True,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        render_mode: Optional[RenderMode] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> TimeMapWKTResponse:
        resp = await send_post_async(
            TimeMapWKTResponse,
            v4_endpoint_path or "time-map/fast",
            self._headers(AcceptType.WKT_NO_HOLES),
            create_time_map_fast_wkt(
                coordinates,
                transportation,
                travel_time,
                level_of_detail,
                snapping,
                polygons_filter,
                render_mode,
                one_to_many,
            ),
            self._sdk_params,
        )
        return resp

    async def h3_fast_async(
        self,
        coordinates: List[Union[Coordinates, H3Centroid]],
        transportation: h3_fast.Transportation,
        properties: List[CellProperty],
        resolution: int,
        travel_time: int = 3600,
        one_to_many: bool = True,
        snapping: Optional[Snapping] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[H3Result]:
        resp = await send_post_async(
            H3Response,
            v4_endpoint_path or "h3/fast",
            self._headers(AcceptType.JSON),
            create_h3_fast(
                coordinates=coordinates,
                transportation=transportation,
                properties=properties,
                resolution=resolution,
                travel_time=travel_time,
                snapping=snapping,
                one_to_many=one_to_many,
            ),
            self._sdk_params,
        )
        return resp.results

    async def geohash_fast_async(
        self,
        coordinates: List[Union[Coordinates, GeohashCentroid]],
        transportation: geohash_fast.Transportation,
        properties: List[CellProperty],
        resolution: int,
        travel_time: int = 3600,
        one_to_many: bool = True,
        snapping: Optional[Snapping] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[GeohashResult]:
        resp = await send_post_async(
            GeohashResponse,
            v4_endpoint_path or "geohash/fast",
            self._headers(AcceptType.JSON),
            create_geohash_fast(
                coordinates=coordinates,
                transportation=transportation,
                properties=properties,
                resolution=resolution,
                travel_time=travel_time,
                snapping=snapping,
                one_to_many=one_to_many,
            ),
            self._sdk_params,
        )
        return resp.results

    async def time_filter_fast_async(
        self,
        locations: List[Location],
        search_ids: Dict[str, List[str]],
        transportation: Transportation,
        travel_time: int = 3600,
        properties: Optional[List[Property]] = None,
        one_to_many: bool = True,
        snapping: Optional[Snapping] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[TimeFilterFastResult]:
        resp = await send_post_async(
            TimeFilterFastResponse,
            v4_endpoint_path or "time-filter/fast",
            self._headers(AcceptType.JSON),
            create_time_filter_fast(
                locations,
                search_ids,
                transportation,
                travel_time,
                properties,
                one_to_many,
                snapping,
            ),
            self._sdk_params,
        )
        return resp.results

    async def postcodes_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 1800,
        properties: Optional[List[Property]] = None,
        range: Optional[FullRange] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[PostcodesResult]:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            PostcodesResponse,
            v4_endpoint_path or "time-filter/postcodes",
            self._headers(AcceptType.JSON),
            create_postcodes(
                coordinates,
                time_info,
                transportation,
                travel_time,
                properties,
                range,
            ),
            self._sdk_params,
        )
        return resp.results

    async def postcodes_districts_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        travel_time: int = 1800,
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        reachable_postcodes_threshold=0.1,
        properties: Optional[List[ZonesProperty]] = None,
        range: Optional[FullRange] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[PostcodesDistrictsResult]:
        time_info = get_time_info(departure_time, arrival_time)

        res = await send_post_async(
            PostcodesDistrictsResponse,
            v4_endpoint_path or "time-filter/postcode-districts",
            self._headers(AcceptType.JSON),
            create_districts(
                coordinates,
                transportation,
                travel_time,
                time_info,
                reachable_postcodes_threshold,
                properties,
                range,
            ),
            self._sdk_params,
        )
        return res.results

    async def postcodes_sectors_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        travel_time: int = 1800,
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        reachable_postcodes_threshold=0.1,
        properties: Optional[List[ZonesProperty]] = None,
        range: Optional[FullRange] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[PostcodesSectorsResult]:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            PostcodesSectorsResponse,
            v4_endpoint_path or "time-filter/postcode-sectors",
            self._headers(AcceptType.JSON),
            create_sectors(
                coordinates,
                transportation,
                travel_time,
                time_info,
                reachable_postcodes_threshold,
                properties,
                range,
            ),
            self._sdk_params,
        )
        return resp.results

    async def routes_async(
        self,
        locations: List[Location],
        search_ids: Dict[str, List[str]],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        properties: Optional[List[Property]] = None,
        range: Optional[FullRange] = None,
        snapping: Optional[Snapping] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[RoutesResult]:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            RoutesResponse,
            v4_endpoint_path or "routes",
            self._headers(AcceptType.JSON),
            create_routes(
                locations,
                search_ids,
                transportation,
                time_info,
                properties,
                range,
                snapping,
            ),
            self._sdk_params,
        )
        return resp.results

    async def time_filter_proto_async(
        self,
        origin: Coordinates,
        destinations: List[Coordinates],
        country: ProtoCountry,
        transportation: Union[
            ProtoTransportation,
            PublicTransportWithDetails,
            DrivingAndPublicTransportWithDetails,
        ],
        travel_time: int,
        one_to_many: bool = True,
        properties: Optional[List[PropertyProto]] = None,
    ) -> TimeFilterProtoResponse:
        if isinstance(transportation, ProtoTransportation):
            transportationMode = transportation.value.name
        else:
            transportationMode = transportation.TYPE.value.name

        resp = await send_proto_async(
            f"https://{self._sdk_params.proto_host}/api/v2/{country.value}/time-filter/fast/{transportationMode}",
            # noqa
            self._proto_headers(),
            create_proto_request(
                origin,
                destinations,
                transportation,
                properties,
                travel_time,
                one_to_many,
            ),
            self._app_id,
            self._api_key,
            self._sdk_params.timeout,
        )
        return resp

    async def time_map_intersection_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        remove_water_bodies: Optional[bool] = None,
        render_mode: Optional[RenderMode] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> TimeMapResult:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            TimeMapResponse,
            v4_endpoint_path or "time-map",
            self._headers(AcceptType.JSON),
            create_time_map_intersection(
                coordinates,
                transportation,
                travel_time,
                time_info,
                search_range,
                level_of_detail,
                snapping,
                polygons_filter,
                remove_water_bodies,
                render_mode,
            ),
            self._sdk_params,
        )
        return resp.results[0]

    # intersection_async was renamed to time_map_intersection_async. Keeping this for legacy users
    async def intersection_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        remove_water_bodies: Optional[bool] = None,
        render_mode: Optional[RenderMode] = None,
    ) -> TimeMapResult:
        resp = await self.time_map_intersection_async(
            coordinates=coordinates,
            transportation=transportation,
            departure_time=departure_time,
            arrival_time=arrival_time,
            travel_time=travel_time,
            search_range=search_range,
            level_of_detail=level_of_detail,
            snapping=snapping,
            polygons_filter=polygons_filter,
            remove_water_bodies=remove_water_bodies,
            render_mode=render_mode,
        )
        return resp

    async def time_map_union_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        remove_water_bodies: Optional[bool] = None,
        render_mode: Optional[RenderMode] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> TimeMapResult:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            TimeMapResponse,
            v4_endpoint_path or "time-map",
            self._headers(AcceptType.JSON),
            create_time_map_union(
                coordinates,
                transportation,
                travel_time,
                time_info,
                search_range,
                level_of_detail,
                snapping,
                polygons_filter,
                remove_water_bodies,
                render_mode,
            ),
            self._sdk_params,
        )

        return resp.results[0]

    # union_async was renamed to time_map_union_async. Keeping this for legacy users
    async def union_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        remove_water_bodies: Optional[bool] = None,
        render_mode: Optional[RenderMode] = None,
    ) -> TimeMapResult:
        resp = await self.time_map_union_async(
            coordinates=coordinates,
            transportation=transportation,
            departure_time=departure_time,
            arrival_time=arrival_time,
            travel_time=travel_time,
            search_range=search_range,
            level_of_detail=level_of_detail,
            snapping=snapping,
            polygons_filter=polygons_filter,
            remove_water_bodies=remove_water_bodies,
            render_mode=render_mode,
        )
        return resp

    async def time_map_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        remove_water_bodies: Optional[bool] = None,
        render_mode: Optional[RenderMode] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[TimeMapResult]:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            TimeMapResponse,
            v4_endpoint_path or "time-map",
            self._headers(AcceptType.JSON),
            create_time_map(
                coordinates,
                transportation,
                travel_time,
                time_info,
                search_range,
                level_of_detail,
                snapping,
                polygons_filter,
                remove_water_bodies,
                render_mode,
            ),
            self._sdk_params,
        )
        return resp.results

    async def time_map_geojson_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        remove_water_bodies: Optional[bool] = None,
        render_mode: Optional[RenderMode] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> FeatureCollection:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            FeatureCollection,
            v4_endpoint_path or "time-map",
            self._headers(AcceptType.GEO_JSON),
            create_time_map_geojson(
                coordinates,
                transportation,
                travel_time,
                time_info,
                search_range,
                level_of_detail,
                snapping,
                polygons_filter,
                remove_water_bodies,
                render_mode,
            ),
            self._sdk_params,
        )
        return resp

    async def time_map_wkt_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        remove_water_bodies: Optional[bool] = None,
        render_mode: Optional[RenderMode] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> TimeMapWKTResponse:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            TimeMapWKTResponse,
            v4_endpoint_path or "time-map",
            self._headers(AcceptType.WKT),
            create_time_map_wkt(
                coordinates,
                transportation,
                travel_time,
                time_info,
                search_range,
                level_of_detail,
                snapping,
                polygons_filter,
                remove_water_bodies,
                render_mode,
            ),
            self._sdk_params,
        )
        return resp

    async def time_map_wkt_no_holes_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        remove_water_bodies: Optional[bool] = None,
        render_mode: Optional[RenderMode] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> TimeMapWKTResponse:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            TimeMapWKTResponse,
            v4_endpoint_path or "time-map",
            self._headers(AcceptType.WKT_NO_HOLES),
            create_time_map_wkt(
                coordinates,
                transportation,
                travel_time,
                time_info,
                search_range,
                level_of_detail,
                snapping,
                polygons_filter,
                remove_water_bodies,
                render_mode,
            ),
            self._sdk_params,
        )
        return resp

    async def h3_intersection_async(
        self,
        coordinates: List[Union[Coordinates, H3Centroid]],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        resolution: int,
        properties: List[CellProperty] = [],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        snapping: Optional[Snapping] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> H3Result:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            H3Response,
            v4_endpoint_path or "h3",
            self._headers(AcceptType.JSON),
            create_h3_intersection(
                coordinates=coordinates,
                transportation=transportation,
                resolution=resolution,
                properties=properties,
                travel_time=travel_time,
                time_info=time_info,
                search_range=search_range,
                snapping=snapping,
            ),
            self._sdk_params,
        )
        return resp.results[0]

    async def h3_union_async(
        self,
        coordinates: List[Union[Coordinates, H3Centroid]],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        resolution: int,
        properties: List[CellProperty] = [],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        snapping: Optional[Snapping] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> H3Result:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            H3Response,
            v4_endpoint_path or "h3",
            self._headers(AcceptType.JSON),
            create_h3_union(
                coordinates=coordinates,
                transportation=transportation,
                resolution=resolution,
                properties=properties,
                travel_time=travel_time,
                time_info=time_info,
                search_range=search_range,
                snapping=snapping,
            ),
            self._sdk_params,
        )

        return resp.results[0]

    async def h3_async(
        self,
        coordinates: List[Union[Coordinates, H3Centroid]],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        resolution: int,
        properties: List[CellProperty] = [],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        snapping: Optional[Snapping] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[H3Result]:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            H3Response,
            v4_endpoint_path or "h3",
            self._headers(AcceptType.JSON),
            create_h3(
                coordinates=coordinates,
                transportation=transportation,
                resolution=resolution,
                properties=properties,
                travel_time=travel_time,
                time_info=time_info,
                search_range=search_range,
                snapping=snapping,
            ),
            self._sdk_params,
        )
        return resp.results

    async def geohash_intersection_async(
        self,
        coordinates: List[Union[Coordinates, GeohashCentroid]],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        resolution: int,
        properties: List[CellProperty] = [],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        snapping: Optional[Snapping] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> GeohashResult:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            GeohashResponse,
            v4_endpoint_path or "geohash",
            self._headers(AcceptType.JSON),
            create_geohash_intersection(
                coordinates=coordinates,
                transportation=transportation,
                resolution=resolution,
                properties=properties,
                travel_time=travel_time,
                time_info=time_info,
                search_range=search_range,
                snapping=snapping,
            ),
            self._sdk_params,
        )
        return resp.results[0]

    async def geohash_union_async(
        self,
        coordinates: List[Union[Coordinates, GeohashCentroid]],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        resolution: int,
        properties: List[CellProperty] = [],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        snapping: Optional[Snapping] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> GeohashResult:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            GeohashResponse,
            v4_endpoint_path or "geohash",
            self._headers(AcceptType.JSON),
            create_geohash_union(
                coordinates=coordinates,
                transportation=transportation,
                resolution=resolution,
                properties=properties,
                travel_time=travel_time,
                time_info=time_info,
                search_range=search_range,
                snapping=snapping,
            ),
            self._sdk_params,
        )

        return resp.results[0]

    async def geohash_async(
        self,
        coordinates: List[Union[Coordinates, GeohashCentroid]],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        resolution: int,
        properties: List[CellProperty] = [],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_time: int = 3600,
        search_range: Optional[Range] = None,
        snapping: Optional[Snapping] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[GeohashResult]:
        time_info = get_time_info(departure_time, arrival_time)

        resp = await send_post_async(
            GeohashResponse,
            v4_endpoint_path or "geohash",
            self._headers(AcceptType.JSON),
            create_geohash(
                coordinates=coordinates,
                transportation=transportation,
                resolution=resolution,
                properties=properties,
                travel_time=travel_time,
                time_info=time_info,
                search_range=search_range,
                snapping=snapping,
            ),
            self._sdk_params,
        )
        return resp.results

    async def distance_map_async(
        self,
        coordinates: List[Coordinates],
        transportation: Union[
            PublicTransport,
            Driving,
            Ferry,
            Walking,
            Cycling,
            DrivingTrain,
            CyclingPublicTransport,
        ],
        departure_time: Optional[datetime] = None,
        arrival_time: Optional[datetime] = None,
        travel_distance: int = 900,
        level_of_detail: Optional[LevelOfDetail] = None,
        snapping: Optional[Snapping] = None,
        polygons_filter: Optional[PolygonsFilter] = None,
        no_holes: Optional[bool] = None,
        v4_endpoint_path: Optional[str] = None,
    ) -> List[TimeMapResult]:
        time_info = get_time_info(departure_time, arrival_time)
        resp = await send_post_async(
            TimeMapResponse,
            v4_endpoint_path or "distance-map",
            self._headers(AcceptType.JSON),
            create_distance_map(
                coordinates,
                transportation,
                travel_distance,
                time_info,
                level_of_detail,
                snapping,
                polygons_filter,
                no_holes,
            ),
            self._sdk_params,
        )
        return resp.results

    @staticmethod
    def _geocoding_reverse_params(lat: float, lng: float) -> Dict[str, str]:
        full_query = {"lat": lat, "lng": lng}
        return {
            key: str(value) for (key, value) in full_query.items() if value is not None
        }

    @staticmethod
    def _geocoding_params(
        query: str,
        limit: Optional[int] = None,
        within_countries: Optional[List[str]] = None,
        format_name: Optional[bool] = None,
        format_exclude_country: Optional[bool] = None,
        bounds: Optional[Rectangle] = None,
    ) -> Dict[str, str]:
        full_query = {
            "query": query,
            "limit": limit,
            "within.country": join_opt(within_countries, ","),
            "format.name": format_name,
            "format.exclude.country": format_exclude_country,
            "bounds": bounds.to_str() if bounds is not None else bounds,
        }
        return {
            key: str(value) for (key, value) in full_query.items() if value is not None
        }

    @staticmethod
    def _proto_headers() -> Dict[str, str]:
        return {
            "Content-Type": AcceptType.OCTET_STREAM.value,
            "User-Agent": f"Travel Time Python SDK {__version__}",
        }

    def _headers(self, accept_type: AcceptType) -> Dict[str, str]:
        return {
            "X-Application-Id": self._app_id,
            "X-Api-Key": self._api_key,
            "User-Agent": self._user_agent,
            "Content-Type": "application/json",
            "Accept": accept_type.value,
        }


def get_time_info(departure_time: Optional[datetime], arrival_time: Optional[datetime]):
    if not departure_time and not arrival_time:
        raise ApiError("either arrival_time or departure_time has to be specified")

    if departure_time and arrival_time:
        raise ApiError("arrival_time and departure_time cannot be both specified")

    if departure_time:
        return DepartureTime(departure_time)
    elif arrival_time:
        return ArrivalTime(arrival_time)



from typing import List, Optional
import typing

from pydantic import BaseModel

from traveltimepy.dto.common import (
    CellProperty,
    Coordinates,
    H3Centroid,
    Snapping,
)
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.responses.h3 import H3Response
from traveltimepy.itertools import split, flatten
from traveltimepy.dto.requests.time_filter_fast import Transportation


class Search(BaseModel):
    id: str
    coords: typing.Union[Coordinates, H3Centroid]
    transportation: Transportation
    travel_time: int
    arrival_time_period: str
    snapping: Optional[Snapping]


class ArrivalSearches(BaseModel):
    many_to_one: List[Search]
    one_to_many: List[Search]


class H3FastRequest(TravelTimeRequest[H3Response]):
    resolution: int
    properties: List[CellProperty]
    arrival_searches: ArrivalSearches

    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        return [
            H3FastRequest(
                resolution=self.resolution,
                properties=self.properties,
                arrival_searches=ArrivalSearches(
                    one_to_many=one_to_many, many_to_one=many_to_one
                ),
            )
            for one_to_many, many_to_one in split(
                self.arrival_searches.one_to_many,
                self.arrival_searches.many_to_one,
                window_size,
            )
        ]

    def merge(self, responses: List[H3Response]) -> H3Response:
        return H3Response(results=flatten([response.results for response in responses]))



class InvalidWKTStringError(Exception):
    def __init__(self, wkt_str):
        super().__init__(f"Invalid WKT string: {wkt_str}")


class InvalidFunctionError(Exception):
    def __init__(self, geometry):
        super().__init__(f"No function found for geometry type: {type(geometry)}")


class InvalidGeometryTypeError(Exception):
    def __init__(self, geometry):
        super().__init__(f"Invalid or unsupported geometry type: {type(geometry)}")


class NullGeometryError(Exception):
    def __init__(self):
        super().__init__("Null, undefined or empty geometry returned")



"""Python sdk for working with traveltime api"""

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    pass

from traveltimepy.dto.transportation import (
    PublicTransport,
    Driving,
    Ferry,
    Walking,
    Cycling,
    DrivingTrain,
    CyclingPublicTransport,
    MaxChanges,
)
from traveltimepy.dto.requests.time_filter_proto import (
    ProtoTransportation,
    ProtoCountry,
)
from traveltimepy.dto.common import (
    Coordinates,
    H3Centroid,
    GeohashCentroid,
    Location,
    Property,
    CellProperty,
    Snapping,
    PropertyProto,
    FullRange,
    Range,
    Rectangle,
    LevelOfDetail,
    DrivingTrafficModel,
)

from traveltimepy.sdk import TravelTimeSdk
from traveltimepy.dto.requests.time_filter_fast import Transportation
from traveltimepy.dto.requests.postcodes_zones import ZonesProperty

__all__ = [
    "__version__",
    "PublicTransport",
    "Driving",
    "Ferry",
    "Walking",
    "Cycling",
    "DrivingTrain",
    "CyclingPublicTransport",
    "MaxChanges",
    "ProtoTransportation",
    "ProtoCountry",
    "Coordinates",
    "H3Centroid",
    "GeohashCentroid",
    "Snapping",
    "Location",
    "Property",
    "CellProperty",
    "PropertyProto",
    "FullRange",
    "Range",
    "Rectangle",
    "LevelOfDetail",
    "TravelTimeSdk",
    "Transportation",
    "ZonesProperty",
    "DrivingTrafficModel",
]



from datetime import datetime
from typing import List, Optional

from pydantic.main import BaseModel


class PublicTransport(BaseModel):
    date_start: datetime
    date_end: datetime


class Features(BaseModel):
    fares: bool
    postcodes: bool
    public_transport: Optional[PublicTransport] = None


class Map(BaseModel):
    name: str
    features: Features


class MapInfoResponse(BaseModel):
    maps: List[Map]



import pytest
from datetime import datetime

from traveltimepy import Coordinates, Driving, Range, TravelTimeSdk
from traveltimepy.dto.common import CellProperty, GeohashCentroid


@pytest.mark.asyncio
async def test_departures(sdk: TravelTimeSdk):
    results = await sdk.geohash_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            GeohashCentroid(geohash_centroid="gcpvhb"),
        ],
        resolution=6,
        properties=[CellProperty.MIN, CellProperty.MAX, CellProperty.MEAN],
        departure_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
    )
    assert len(results) == 2


@pytest.mark.asyncio
async def test_arrivals(sdk: TravelTimeSdk):
    results = await sdk.geohash_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            GeohashCentroid(geohash_centroid="gcpvhb"),
        ],
        resolution=6,
        properties=[CellProperty.MIN, CellProperty.MAX, CellProperty.MEAN],
        arrival_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
    )
    assert len(results) == 2


@pytest.mark.asyncio
async def test_union_departures(sdk: TravelTimeSdk):
    result = await sdk.geohash_union_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            GeohashCentroid(geohash_centroid="gcpvhb"),
        ],
        resolution=6,
        properties=[CellProperty.MIN, CellProperty.MAX, CellProperty.MEAN],
        departure_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
    )
    assert len(result.cells) > 0


@pytest.mark.asyncio
async def test_intersection_arrivals(sdk: TravelTimeSdk):
    result = await sdk.geohash_intersection_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            GeohashCentroid(geohash_centroid="gcpvhb"),
        ],
        resolution=6,
        properties=[CellProperty.MIN, CellProperty.MAX, CellProperty.MEAN],
        arrival_time=datetime.now(),
        travel_time=900,
        transportation=Driving(),
        search_range=Range(enabled=True, width=1800),
    )
    assert len(result.cells) > 0



from typing import List

import pytest
from datetime import datetime

from traveltimepy import PublicTransport, Driving, Location, Coordinates
from traveltimepy.dto.common import Snapping, SnappingAcceptRoads, SnappingPenalty
from traveltimepy.sdk import TravelTimeSdk


@pytest.mark.asyncio
async def test_departures(sdk: TravelTimeSdk, locations):
    results = await sdk.routes_async(
        locations=locations,
        search_ids={
            "London center": ["Hyde Park", "ZSL London Zoo"],
            "ZSL London Zoo": ["Hyde Park", "London center"],
        },
        transportation=PublicTransport(),
        departure_time=datetime.now(),
    )
    assert len(results) == 2


@pytest.mark.asyncio
async def test_arrivals(sdk: TravelTimeSdk, locations):
    results = await sdk.routes_async(
        locations=locations,
        search_ids={
            "London center": ["Hyde Park", "ZSL London Zoo"],
            "ZSL London Zoo": ["Hyde Park", "London center"],
        },
        transportation=PublicTransport(),
        departure_time=datetime.now(),
    )
    assert len(results) == 2


@pytest.mark.asyncio
async def test_snapping(sdk: TravelTimeSdk):
    locations: List[Location] = [
        Location(id="A", coords=Coordinates(lat=53.806479, lng=-2.615711)),
        Location(id="B", coords=Coordinates(lat=53.810129, lng=-2.601099)),
    ]
    result_with_penalty = await sdk.routes_async(
        locations=locations,
        search_ids={
            "A": ["B"],
        },
        transportation=Driving(),
        departure_time=datetime.now(),
    )
    traveltime_with_penalty = (
        result_with_penalty[0].locations[0].properties[0].travel_time
    )
    result_without_penalty = await sdk.routes_async(
        locations=locations,
        search_ids={
            "A": ["B"],
        },
        transportation=Driving(),
        departure_time=datetime.now(),
        snapping=Snapping(
            penalty=SnappingPenalty.DISABLED,
            accept_roads=SnappingAcceptRoads.ANY_DRIVABLE,
        ),
    )
    traveltime_without_penalty = (
        result_without_penalty[0].locations[0].properties[0].travel_time
    )
    assert traveltime_with_penalty is not None
    assert traveltime_without_penalty is not None
    assert traveltime_with_penalty > traveltime_without_penalty



import pytest

from traveltimepy import Transportation, TravelTimeSdk
from traveltimepy.dto.common import CellProperty, Coordinates, H3Centroid


@pytest.mark.asyncio
async def test_one_to_many(sdk: TravelTimeSdk):
    results = await sdk.h3_fast_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            H3Centroid(h3_centroid="87195da49ffffff"),
        ],
        resolution=7,
        properties=[CellProperty.MIN, CellProperty.MAX, CellProperty.MEAN],
        travel_time=900,
        transportation=Transportation(type="public_transport"),
    )

    assert len(results) == 2


@pytest.mark.asyncio
async def test_many_to_one(sdk: TravelTimeSdk):
    results = await sdk.h3_fast_async(
        coordinates=[
            Coordinates(lat=51.507609, lng=-0.128315),
            H3Centroid(h3_centroid="87195da49ffffff"),
        ],
        resolution=7,
        properties=[CellProperty.MIN, CellProperty.MAX, CellProperty.MEAN],
        travel_time=900,
        transportation=Transportation(type="public_transport"),
        one_to_many=False,
    )

    assert len(results) == 2



from typing import List, Optional

from pydantic import BaseModel


class Property(BaseModel):
    travel_time: Optional[int] = None
    distance: Optional[int] = None


class Postcode(BaseModel):
    code: str
    properties: List[Property]


class PostcodesResult(BaseModel):
    search_id: str
    postcodes: List[Postcode]


class PostcodesResponse(BaseModel):
    results: List[PostcodesResult]



from datetime import datetime
from enum import Enum
from typing import List, Union, Optional

from pydantic import BaseModel

from traveltimepy.dto.common import Coordinates, FullRange
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.responses.zones import (
    PostcodesDistrictsResponse,
    PostcodesSectorsResponse,
)
from traveltimepy.itertools import split, flatten
from traveltimepy.dto.transportation import (
    PublicTransport,
    Driving,
    Ferry,
    Walking,
    Cycling,
    DrivingTrain,
    CyclingPublicTransport,
)


class ZonesProperty(str, Enum):
    TRAVEL_TIME_REACHABLE = "travel_time_reachable"
    TRAVEL_TIME_ALL = "travel_time_all"
    COVERAGE = "coverage"


class ArrivalSearch(BaseModel):
    id: str
    coords: Coordinates
    travel_time: int
    arrival_time: datetime
    reachable_postcodes_threshold: float
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ]
    properties: List[ZonesProperty]
    range: Optional[FullRange]


class DepartureSearch(BaseModel):
    id: str
    coords: Coordinates
    travel_time: int
    departure_time: datetime
    reachable_postcodes_threshold: float
    transportation: Union[
        PublicTransport,
        Driving,
        Ferry,
        Walking,
        Cycling,
        DrivingTrain,
        CyclingPublicTransport,
    ]
    properties: List[ZonesProperty]
    range: Optional[FullRange]


class PostcodesSectorsRequest(TravelTimeRequest[PostcodesSectorsResponse]):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]

    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        return [
            PostcodesSectorsRequest(
                departure_searches=departures, arrival_searches=arrivals
            )
            for departures, arrivals in split(
                self.departure_searches, self.arrival_searches, window_size
            )
        ]

    def merge(
        self, responses: List[PostcodesSectorsResponse]
    ) -> PostcodesSectorsResponse:
        return PostcodesSectorsResponse(
            results=sorted(
                flatten([response.results for response in responses]),
                key=lambda res: res.search_id,
            )
        )


class PostcodesDistrictsRequest(TravelTimeRequest[PostcodesDistrictsResponse]):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]

    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        return [
            PostcodesDistrictsRequest(
                departure_searches=departures, arrival_searches=arrivals
            )
            for departures, arrivals in split(
                self.departure_searches, self.arrival_searches, window_size
            )
        ]

    def merge(
        self, responses: List[PostcodesDistrictsResponse]
    ) -> PostcodesDistrictsResponse:
        return PostcodesDistrictsResponse(
            results=sorted(
                flatten([response.results for response in responses]),
                key=lambda res: res.search_id,
            )
        )



from typing import List
from traveltimepy.dto.requests.request import TravelTimeRequest
from traveltimepy.dto.requests.time_map import DepartureSearch, ArrivalSearch
from traveltimepy.dto.responses.time_map_wkt import TimeMapWKTResponse
from traveltimepy.itertools import split, flatten


class TimeMapWKTRequest(TravelTimeRequest[TimeMapWKTResponse]):
    departure_searches: List[DepartureSearch]
    arrival_searches: List[ArrivalSearch]

    def split_searches(self, window_size: int) -> List[TravelTimeRequest]:
        return [
            TimeMapWKTRequest(departure_searches=departures, arrival_searches=arrivals)
            for departures, arrivals in split(
                self.departure_searches, self.arrival_searches, window_size
            )
        ]

    def merge(self, responses: List[TimeMapWKTResponse]) -> TimeMapWKTResponse:
        return TimeMapWKTResponse(
            results=sorted(
                flatten([response.results for response in responses]),
                key=lambda res: res.search_id,
            )
        )



from datetime import datetime, time
from enum import Enum
from typing import List, Union, Optional

from pydantic import field_validator
from typing_extensions import Literal
from pydantic.main import BaseModel


class Coordinates(BaseModel):
    lat: float
    lng: float

    @field_validator("lat")
    @classmethod
    def validate_latitude(cls, v):
        if not -90 <= v <= 90:
            raise ValueError("Latitude must be between -90 and 90.")
        return v

    @field_validator("lng")
    @classmethod
    def validate_longitude(cls, v):
        if not -180 <= v <= 180:
            raise ValueError("Longitude must be between -180 and 180.")
        return v


class GeohashCentroid(BaseModel):
    geohash_centroid: str


class H3Centroid(BaseModel):
    h3_centroid: str


class Location(BaseModel):
    id: str
    coords: Coordinates

    def __hash__(self):
        return hash(self.id)


class BasicPart(BaseModel):
    id: int
    mode: str
    directions: str
    distance: int
    travel_time: int
    coords: List[Coordinates]
    type: Literal["basic"]


class RoadPart(BaseModel):
    id: int
    mode: str
    directions: str
    distance: int
    travel_time: int
    coords: List[Coordinates]
    type: Literal["road"]
    road: Optional[str] = None
    turn: Optional[str] = None


class StartEndPart(BaseModel):
    id: int
    mode: str
    directions: str
    distance: int
    travel_time: int
    coords: List[Coordinates]
    type: Literal["start_end"]
    direction: str


class PublicTransportPart(BaseModel):
    id: int
    mode: str
    directions: str
    distance: int
    travel_time: int
    coords: List[Coordinates]
    line: str
    departure_station: str
    arrival_station: str
    departs_at: time
    arrives_at: time
    num_stops: int
    type: Literal["public_transport"]


class Route(BaseModel):
    departure_time: datetime
    arrival_time: datetime
    parts: List[Union[BasicPart, PublicTransportPart, StartEndPart, RoadPart]]


class Ticket(BaseModel):
    type: str
    price: float
    currency: str


class FareBreakdown(BaseModel):
    modes: List[str]
    route_part_ids: List[int]
    tickets: List[Ticket]


class Fares(BaseModel):
    breakdown: List[FareBreakdown]
    tickets_total: List[Ticket]


class Rectangle(BaseModel):
    min_lat: float
    max_lat: float
    min_lng: float
    max_lng: float

    def to_str(self):
        return f"{self.min_lat},{self.min_lng},{self.max_lat},{self.max_lng}"


class Property(str, Enum):
    TRAVEL_TIME = "travel_time"
    DISTANCE = "distance"
    ROUTE = "route"
    FARES = "fares"


class CellProperty(str, Enum):
    MIN = "min"
    MAX = "max"
    MEAN = "mean"


class SnappingPenalty(str, Enum):
    ENABLED = "enabled"
    DISABLED = "disabled"


class SnappingAcceptRoads(str, Enum):
    BOTH_DRIVABLE_AND_WALKABLE = "both_drivable_and_walkable"
    ANY_DRIVABLE = "any_drivable"


class DrivingTrafficModel(str, Enum):
    OPTIMISTIC = "optimistic"
    BALANCED = "balanced"
    PESSIMISTIC = "pessimistic"


class Snapping(BaseModel):
    penalty: Optional[SnappingPenalty] = SnappingPenalty.ENABLED
    accept_roads: Optional[SnappingAcceptRoads] = (
        SnappingAcceptRoads.BOTH_DRIVABLE_AND_WALKABLE
    )


class PropertyProto(int, Enum):
    DISTANCE = 1


class FullRange(BaseModel):
    enabled: bool
    max_results: int
    width: int


class Range(BaseModel):
    enabled: bool
    width: int


class LevelOfDetail(BaseModel):
    scale_type: Literal["simple", "simple_numeric", "coarse_grid"] = "simple"
    level: Optional[Union[int, str]] = None
    square_size: Optional[int] = None


class PolygonsFilter(BaseModel):
    limit: int


class RenderMode(str, Enum):
    APPROXIMATE_TIME_FILTER = "approximate_time_filter"
    ROAD_BUFFERING = "road_buffering"


class TimeInfo:
    def __init__(self, time_value: datetime):
        self.value = time_value


class DepartureTime(TimeInfo):
    pass


class ArrivalTime(TimeInfo):
    pass
